/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "kOmegaSST.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace turbulenceModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(kOmegaSST, 0);
addToRunTimeSelectionTable(turbulenceModel, kOmegaSST, dictionary);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

// from components
kOmegaSST::kOmegaSST
(
    const volVectorField& U,
    const surfaceScalarField& phi,
    transportModel& lamTransportModel
)
:
    turbulenceModel(typeName, U, phi, lamTransportModel),

    Cmu(turbulenceModelCoeffs_.lookup("Cmu")),
    C1_om(turbulenceModelCoeffs_.lookup("C1_om")),    
    C1_eps(turbulenceModelCoeffs_.lookup("C1_eps")),
    C2_eps(turbulenceModelCoeffs_.lookup("C2_eps")),
    C2_om(turbulenceModelCoeffs_.lookup("C2_om")),
    alphaEps(turbulenceModelCoeffs_.lookup("alphaEps")),
    alphaOmega(turbulenceModelCoeffs_.lookup("alphaOmega")),
    alphaKEps(turbulenceModelCoeffs_.lookup("alphaKEps")),
    alphaKOmega(turbulenceModelCoeffs_.lookup("alphaKOmega")),
    a1(turbulenceModelCoeffs_.lookup("a1")),   

    k_
    (
        IOobject
        (
            "k",
            runTime_.timeName(),
            mesh_,
            IOobject::MUST_READ,
            IOobject::AUTO_WRITE
        ),
        mesh_
    ),

    omega_
    (
        IOobject
        (
            "omega",
            runTime_.timeName(),
            mesh_,
            IOobject::MUST_READ,
            IOobject::AUTO_WRITE
        ),
        mesh_
    ),

    y_(mesh_),


    // just for initialization

    nut_(k_/omega_),

    alpha(k_/k_),

    alphaK(k_/k_)

{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

tmp<volTensorField> kOmegaSST::R() const
{
    return tmp<volTensorField>
    (
        new volTensorField
        (
            IOobject
            (
                "R",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            ((2.0/3.0)*I)*k_ - nut_*2*symm(fvc::grad(U_)),
            k_.boundaryField().types()
        )
    );
}


tmp<fvVectorMatrix> kOmegaSST::divR(volVectorField& U) const
{
    return
    (
      - fvm::laplacian(nuEff(), U)
      - fvc::div(nuEff()*dev(fvc::grad(U)().T()))
    );
}


bool kOmegaSST::read()
{
    if (turbulenceModel::read())
    {
        turbulenceModelCoeffs_.lookup("Cmu") >> Cmu;
        turbulenceModelCoeffs_.lookup("C1_eps") >> C1_eps;
        turbulenceModelCoeffs_.lookup("C2_eps") >> C2_eps;
        turbulenceModelCoeffs_.lookup("C1_om") >> C1_om;
        turbulenceModelCoeffs_.lookup("C2_om") >> C2_om;
        turbulenceModelCoeffs_.lookup("alphaEps") >> alphaEps;
        turbulenceModelCoeffs_.lookup("alphaOmega") >> alphaOmega;
        turbulenceModelCoeffs_.lookup("alphaKEps") >> alphaKEps;
        turbulenceModelCoeffs_.lookup("alphaKOmega") >> alphaKOmega;
        turbulenceModelCoeffs_.lookup("a1") >> a1;

        return true;
    }
    else
    {
        return false;
    }
}


void kOmegaSST::correct()
{
    transportModel_.correct();

    if (!turbulence_)
    {
        return;
    }

    turbulenceModel::correct();

    if (mesh_.moving())
    {
        y_.correct();
    }

    volScalarField G = nut_*2*magSqr(symm(fvc::grad(U_)));

    // Calculate parameters and coefficients for low-Reynolds number model

    dimensionedScalar CDkwlim
      (
       "CDkwlim",
       dimensionSet(0, 0, -2, 0, 0),
       10e-10
       );


    volScalarField CDkw = max(2*alphaEps*pow(omega_,-1)
      *( fvc::grad(k_) & fvc::grad(omega_)),CDkwlim);




    volScalarField arg1 = min(max(sqrt(k_)/0.09*pow(omega_,-1)*pow(y_,-1),
				  500*nu()*pow(omega_,-1)*pow(sqr(y_),-1)),
			      4*alphaEps*k_*pow(CDkw,-1)*pow(sqr(y_),-1));

    volScalarField arg2 = max(2*sqrt(k_)/0.09*pow(omega_,-1)*pow(y_,-1),
			      500*nu()*pow(sqr(y_),-1)*pow(omega_,-1));

    
    volScalarField F1 = tanh(pow(arg1,4));

    alpha = F1*alphaOmega + (scalar(1)-F1)*alphaEps;

    alphaK = F1*alphaKOmega + (scalar(1)-F1)*alphaKEps;

    volScalarField F2 = tanh(pow(arg2,2));

    volScalarField C1 = F1*C1_om + (scalar(1)-F1)*C1_eps;
    volScalarField C2 = F1*C2_om + (scalar(1)-F1)*C2_eps;



    // Dissipation equation

      tmp<fvScalarMatrix> omegaEqn
    (
        fvm::ddt(omega_)
      + fvm::div(phi_, omega_)
      - fvm::laplacian(DomegaEff(), omega_)
     ==
        C1*G*omega_/k_
      - fvm::Sp(C2*omega_, omega_)
	+ 2*(scalar(1)-F1)*alphaEps*pow(omega_,-1)
      *( fvc::grad(k_) & fvc::grad(omega_))
    );

    omegaEqn().relax();
    solve(omegaEqn);
    bound(omega_, omega0_);



    // Turbulent kinetic energy equation

    tmp<fvScalarMatrix> kEqn
    (
        fvm::ddt(k_)
      + fvm::div(phi_, k_)
      - fvm::laplacian(DkEff(), k_)
     ==
        G - fvm::Sp(Cmu*omega_, k_)
    );

    kEqn().relax();
    solve(kEqn);
    bound(k_, k0_);


    // Re-calculate viscosity
    nut_ = a1*k_/max(a1*omega_,mag(fvc::curl(U_))*F2);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace turbulenceModels
} // End namespace Foam

// ************************************************************************* //
